#!/usr/bin/env python3
import unittest
import os
import subprocess


test_names = ["test1", "test3"]
test_data = {
    test_names[0]: "data/test_isolate_01.fa",  # Test published resistance
    test_names[1]: "data/test_isolate_03.fa",  # Test no resistance
}
run_test_dir = "unit_tests"
blast_out_dir = "blast_out"

configs = {}


class ResFinderTest(unittest.TestCase):

    def setUp(self):
        # Change working dir to test dir
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        # Does not allow running two tests in parallel
        os.makedirs(run_test_dir, exist_ok=False)
        with open("test_config.txt", "r") as fh:
            for line in fh:
                line = line.rstrip()

                if not(line):
                    continue

                if(line.startswith("#")):
                    continue

                key, path = line.split()
                configs[key] = path

    def test_run_resfinder(self):

        test1_dir = run_test_dir + "/" + test_names[0]
        test1_blast_dir = test1_dir + "/" + blast_out_dir
        os.makedirs(test1_dir)
        os.makedirs(test1_blast_dir)

        cmd = ("python3 ../ResFinder.py"
               + " -i " + test_data[test_names[0]]
               + " -o " + test1_blast_dir
               + " -p " + configs["resfinder_dbs"]
               + " -l 0.6"
               + " -t 0.9")

        print("Run cmd: " + cmd)

        process = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)

        outs = process.stdout.decode()
        errs = process.stderr.decode()

        self.assertNotIn("Error", errs)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

from __future__ import print_function
import argparse
import os.path
import re
import shutil
from signal import *
import tempfile
import sys
import subprocess
from .feature import Feature, ResGene, Mutation, ResMutation
from .abclassdef import ABClassDefinition


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


class PhenoDB(dict):
    """ Loads a text table into dict.
        The dict consists of Phenotype objects. The keys are unique ids.
    """

    def __init__(self, abclassdef_file, acquired_file=None, point_file=None,
                 disclassdef_file=None, disinf_file=None):

        # Stores non-redundant complete list of antibiotics in DB.
        self.antibiotics = {}

        # ID in DB is <gene>_<group>_<acc>. Ex: blaB-2_1_AF189300
        # The <group> should be redundant, as the acc will be enough to
        # define which variant it is.
        # This code uses the 'new' ID <gene>_<acc>, but has this dict so other
        # parts can translate the new ids into the old ones that includes
        # <group>
        self.id_to_idwithvar = {}

        self.ab_class_defs = ABClassDefinition(abclassdef_file)

        if(acquired_file is None and point_file is None
           and disinf_file is None):
            eprint("ERROR: No phenotype database files where specified.")
            quit(1)

        # mut_type_is_defined indicates the new pointfinder db with AA/NUC
        # instead of DNA/RNA
        self.mut_type_is_defined = False

        if(acquired_file):
            self.load_acquired_db(acquired_file)

        if(disinf_file):
            self.dis_class_defs = ABClassDefinition(disclassdef_file)
            self.load_disinfectant_db(disinf_file)

        if(point_file):

            if os.path.basename(point_file) == "resistens-overview.txt":
                self.load_point_old_db(point_file)
            else:
                self.load_point_db(point_file)

        self.unknown_pheno = Phenotype(unique_id="unknown",
                                       antibiotics=[],
                                       sug_phenotype=(),
                                       pub_phenotype=(),
                                       pmid="-",
                                       notes="Phenotype not found in database",
                                       res_db="ResFinder")

    def load_acquired_db(self, txt_file):

        # Test file for illegal encodings
        with open(txt_file, "r") as fh:
            line_counter = 1
            try:
                for line in fh:
                    line_counter += 1
            except UnicodeDecodeError:
                eprint("PhenoDB load_acquired_dbacq UnicodeDecodeError in "
                       "line {0}".format(line_counter))
                eprint("\t\"" + line + "\"")
                sys.exit("UnicodeDecodeError")

        with open(txt_file, "r") as fh:
            # Skip headers
            fh.readline()
            line_counter = 1

            for line in fh:
                if(line.startswith("#")):
                    continue

                try:
                    line_counter += 1

                    # line = line.encode("latin_1")
                    line = line.rstrip()
                    line_list = line.split("\t")
                    line_list = list(map(str.rstrip, line_list))

                    # ID in DB is <gene>_<group>_<acc>. Ex: blaB-2_1_AF189300.
                    # The gene + acc should be unique and is used here.
                    id_with_group = line_list[0]
                    phenodb_id = id_with_group.split("_")

                    # This code is temporary
                    # TODO: Remove when database has been reformatted.
                    if(len(phenodb_id) < 3):
                        unique_id = phenodb_id
                    else:
                        accno = "_".join(phenodb_id[2:])
                        unique_id = "{0}_{1}".format(phenodb_id[0], accno)

                    self.id_to_idwithvar[unique_id] = id_with_group

                    # ab_class = self.get_csv_tuple(line_list[1].lower())

                    pub_phenotype = self.get_csv_tuple(line_list[2].lower())
                    if("unknown" in pub_phenotype or "none" in pub_phenotype):
                        pub_phenotype = ()

                    abs = []
                    for ab_name in pub_phenotype:
                        # TODO: Fix database
                        if(ab_name == "see notes"):
                            continue
                        classes = self.ab_class_defs.get(ab_name, None)
                        if(classes is None):
                            eprint("Entry {id:s} contained antibiotic "
                                   "'{ab:s}' not found in ab class "
                                   "def file."
                                   .format(id=unique_id, ab=ab_name))
                            continue
                        ab = Antibiotics(name=ab_name, classes=classes)
                        abs.append(ab)
                        for class_ in classes:
                            if(class_ in self.antibiotics):
                                self.antibiotics[class_][ab] = True
                            else:
                                self.antibiotics[class_] = {}
                                self.antibiotics[class_][ab] = True

                    # phenotype = list(pub_phenotype)

                    pmid = self.get_csv_tuple(line_list[3].lower())

                    if(len(line_list) > 4 and line_list[4]):
                        res_mechanics = line_list[4].lower()
                    else:
                        res_mechanics = ""

                    if(len(line_list) > 5 and line_list[5]):
                        notes = line_list[5]
                    else:
                        notes = ""

                    # Line list pos 6 is required genes

                    #
                    # The following should be written out
                    #
                    # if(len(line_list) > 6 and line_list[6]):
                    #    sug_phenotype = self.get_csv_tuple(line_list[6])
                    #    for p in sug_phenotype:
                    #        if p not in pub_phenotype:
                    #            phenotype.append(p)
                    # else:
                    sug_phenotype = ()

                    if(len(line_list) > 7 and line_list[7]):
                        gene_class = line_list[7].lower()
                    else:
                        gene_class = None
                    if(len(line_list) > 8 and line_list[8]):
                        susceptibile = self.get_csv_tuple(line_list[6])
                    else:
                        susceptibile = ()
                    if(len(line_list) > 9 and line_list[9]):
                        species = self.get_csv_tuple(line_list[9])
                    else:
                        species = None

                    pheno = Phenotype(unique_id, abs,
                                      sug_phenotype, pub_phenotype, pmid,
                                      susceptibile=susceptibile,
                                      gene_class=gene_class, notes=notes,
                                      res_db="ResFinder")

                    pheno_lst = self.get(unique_id, [])
                    pheno_lst.append(pheno)
                    self[unique_id] = pheno_lst

                except IndexError:
                    eprint("PhenoDB load_acquired_dbacq IndexError in line {0}"
                           .format(line_counter))
                    eprint("Split line:\n{0}".format(str(line_list)))

    def load_point_db(self, txt_file):

        with open(txt_file, "r") as fh:
            # Skip headers
            fh.readline()
            line_counter = 1

            for line in fh:
                if(line.startswith("#")):
                    continue

                try:
                    line_counter += 1

                    line_list = line.split("\t")
                    line_list = list(map(str.rstrip, line_list))
                    # ID in DB is Gene-AAMut-Pos and is not unique
                    gene_id = line_list[0]
                    mut_type = line_list[1]
                    codon_pos = line_list[3]
                    res_codon_str = line_list[6].lower()

                    if mut_type == 'AA' or mut_type == 'NUC':
                        self.mut_type_is_defined = True

                    if not self.mut_type_is_defined:
                        eprint("Warning: Your PointFinder database is not "
                               "up to date. This may effect the results.")
                    # Check if the entry is with a promoter
                    gene_id = PhenoDB.if_promoter_rename(gene_id)

                    gene_name = gene_id.replace("_", ";;")
                    pub_phenotype = self.get_csv_tuple(line_list[8].lower())
                    if("unknown" in pub_phenotype or "none" in pub_phenotype):
                        pub_phenotype = ()

                    pmid = self.get_csv_tuple(line_list[9].lower())

                    # TODO: Remove this tuple and its dependencies.
                    sug_phenotype = ()
                    if self.mut_type_is_defined:
                        unique_id = "%s_%s_%s_%s" % (gene_name, codon_pos,
                                                     res_codon_str, mut_type)
                    else:
                        unique_id = "%s_%s_%s" % (gene_name, codon_pos,
                                                     res_codon_str)

                    abs = []
                    for ab_name in pub_phenotype:
                        # TODO: Fix database
                        if(ab_name == "see notes"):
                            continue
                        classes = self.ab_class_defs.get(ab_name, None)
                        if(classes is None):
                            eprint("Entry {id:s} contained antibiotic "
                                   "'{ab:s}' not found in ab class "
                                   "def file."
                                   .format(id=unique_id, ab=ab_name))
                            continue
                        ab = Antibiotics(name=ab_name, classes=classes)
                        abs.append(ab)
                        for class_ in classes:
                            if(class_ in self.antibiotics):
                                self.antibiotics[class_][ab] = True
                            else:
                                self.antibiotics[class_] = {}
                                self.antibiotics[class_][ab] = True
                    if(len(line_list) > 10 and line_list[10]):
                        res_mechanics = line_list[10]
                    else:
                        res_mechanics = None

                    if(len(line_list) > 11 and line_list[11]):
                        notes = line_list[11]
                    else:
                        notes = ""

                    # Load required mutations.
                    # A mutation can dependent on a group of other
                    # mutations. It is also possible for a mutation to
                    # be dependent on either group A, B, C etc.
                    # The string stored in this field divides each
                    # group using a semicolon, and each mutation using
                    # a comma.
                    # The individual mutation notation is also a little
                    # different from elsewhere, see MutationGenotype
                    # class for more info.
                    # Requied mutations are stored in a tuple of tuples
                    # of MutationGenotypes. Outer tuple seperate
                    # groups, inner tuples seperate mutations.
                    if(len(line_list) > 12 and line_list[12]):
                        mut_groups_str = PhenoDB.get_csv_tuple(line_list[12],
                                                               sep=";",
                                                               lower=False)
                        if(mut_groups_str):
                            mut_groups = []

                            for group in mut_groups_str:
                                group_list = []
                                mut_strings = PhenoDB.get_csv_tuple(
                                    group, lower=False)
                                for mut_str in mut_strings:
                                    mut = MutationGenotype(mut_str)
                                    group_list.append(mut)

                                mut_groups.append(tuple(group_list))

                            mut_groups = tuple(mut_groups)

                        else:
                            mut_groups = None
                    else:
                        mut_groups = None

                    pheno = Phenotype(unique_id, abs,
                                      sug_phenotype, pub_phenotype, pmid,
                                      notes=notes, res_mechanics=res_mechanics,
                                      req_muts=mut_groups, res_db="PointFinder"
                                      )

                    pheno_lst = self.get(unique_id, [])
                    pheno_lst.append(pheno)
                    self[unique_id] = pheno_lst

                    # A pointmutation with several different res codons will
                    # never be found using all the res_codons. Instead it will
                    # be found with just one.
                    # Alternative unique ids are therefore made using just a
                    # single res_codon.
                    res_codon = self.get_csv_tuple(line_list[6].lower())
                    if(len(res_codon) > 1):
                        for codon in res_codon:
                            if self.mut_type_is_defined:
                                unique_id_alt = (gene_name + "_" + codon_pos
                                                 + "_" + codon + "_" + mut_type)
                            else:
                                unique_id_alt = (gene_name + "_" + codon_pos
                                                 + "_" + codon)
                            self[unique_id_alt] = pheno_lst
                except IndexError:
                    eprint("Error in line " + str(line_counter))
                    eprint("Split line:\n" + str(line_list))

    def load_point_old_db(self, txt_file):

        with open(txt_file, "r") as fh:
            # Skip headers
            fh.readline()
            line_counter = 1

            for line in fh:
                if(line.startswith("#")):
                    continue

                try:
                    line_counter += 1

                    line_list = line.split("\t")
                    line_list = list(map(str.rstrip, line_list))

                    # ID in DB is just Gene ID and is not unique
                    phenodb_id = line_list[0]
                    codon_pos = line_list[2]
                    res_codon_str = line_list[5].lower()

                    # Check if the entry is with a promoter
                    phenodb_id = PhenoDB.if_promoter_rename(phenodb_id)

                    unique_id = (phenodb_id + "_" + codon_pos + "_"
                                 + res_codon_str)

                    pub_phenotype = self.get_csv_tuple(line_list[6].lower())
                    if("unknown" in pub_phenotype or "none" in pub_phenotype):
                        pub_phenotype = ()

                    pmid = self.get_csv_tuple(line_list[7].lower())

                    # TODO: Remove this tuple and its dependencies.
                    sug_phenotype = ()

                    abs = []
                    for ab_name in pub_phenotype:
                        # TODO: Fix database
                        if(ab_name == "see notes"):
                            continue
                        classes = self.ab_class_defs.get(ab_name, None)
                        if(classes is None):
                            eprint("Entry {id:s} contained antibiotic "
                                   "'{ab:s}' not found in ab class "
                                   "def file."
                                   .format(id=unique_id, ab=ab_name))
                            continue
                        ab = Antibiotics(name=ab_name, classes=classes)
                        abs.append(ab)
                        for class_ in classes:
                            if(class_ in self.antibiotics):
                                self.antibiotics[class_][ab] = True
                            else:
                                self.antibiotics[class_] = {}
                                self.antibiotics[class_][ab] = True

                    if(len(line_list) > 8 and line_list[8]):
                        res_mechanics = line_list[8]
                    else:
                        res_mechanics = None

                    if(len(line_list) > 9 and line_list[9]):
                        notes = line_list[9]
                    else:
                        notes = ""

                    # Load required mutations.
                    # A mutation can dependent on a group of other
                    # mutations. It is also possible for a mutation to
                    # be dependent on either group A, B, C etc.
                    # The string stored in this field divides each
                    # group using a semicolon, and each mutation using
                    # a comma.
                    # The individual mutation notation is also a little
                    # different from elsewhere, see MutationGenotype
                    # class for more info.
                    # Requied mutations are stored in a tuple of tuples
                    # of MutationGenotypes. Outer tuple seperate
                    # groups, inner tuples seperate mutations.
                    if(len(line_list) > 10 and line_list[10]):
                        mut_groups_str = PhenoDB.get_csv_tuple(line_list[10],
                                                               sep=";",
                                                               lower=False)
                        if(mut_groups_str):
                            mut_groups = []

                            for group in mut_groups_str:
                                group_list = []
                                mut_strings = PhenoDB.get_csv_tuple(
                                    group, lower=False)
                                for mut_str in mut_strings:
                                    mut = MutationGenotype(mut_str)
                                    group_list.append(mut)

                                mut_groups.append(tuple(group_list))

                            mut_groups = tuple(mut_groups)

                        else:
                            mut_groups = None
                    else:
                        mut_groups = None

                    pheno = Phenotype(unique_id, abs,
                                      sug_phenotype, pub_phenotype, pmid,
                                      notes=notes, res_mechanics=res_mechanics,
                                      req_muts=mut_groups, res_db="PointFinder"
                                      )

                    pheno_lst = self.get(unique_id, [])
                    pheno_lst.append(pheno)
                    self[unique_id] = pheno_lst

                    # A pointmutation with several different res codons will
                    # never be found using all the res_codons. Instead it will
                    # be found with just one.
                    # Alternative unique ids are therefore made using just a
                    # single res_codon.
                    res_codon = self.get_csv_tuple(line_list[5].lower())
                    if(len(res_codon) > 1):
                        for codon in res_codon:
                            unique_id_alt = (phenodb_id + "_" + codon_pos
                                             + "_" + codon)
                            self[unique_id_alt] = pheno_lst
                except IndexError:
                    eprint("Error in line " + str(line_counter))
                    eprint("Split line:\n" + str(line_list))

    def load_disinfectant_db(self, txt_file):

        # Test file for illegal encodings
        with open(txt_file, "r") as fh:
            line_counter = 1
            try:
                for line in fh:
                    line_counter += 1
            except UnicodeDecodeError:
                eprint("PhenoDB load_acquired_dbacq UnicodeDecodeError in "
                       "line {0}".format(line_counter))
                eprint("\t\"" + line + "\"")
                sys.exit("UnicodeDecodeError")

        with open(txt_file, "r") as fh:
            # Skip headers
            fh.readline()
            line_counter = 1

            for line in fh:
                if(line.startswith("#")):
                    continue

                try:
                    line_counter += 1

                    # line = line.encode("latin_1")
                    line = line.rstrip()
                    line_list = line.split("\t")
                    line_list = list(map(str.rstrip, line_list))

                    # ID in DB is <gene>_<group>_<acc>. Ex: blaB-2_1_AF189300.
                    # The gene + acc should be unique and is used here.
                    id_with_group = line_list[0]
                    phenodb_id = id_with_group.split("_")

                    # This code is temporary
                    # TODO: Remove when database has been reformatted.
                    if(len(phenodb_id) < 3):
                        unique_id = phenodb_id
                    else:
                        accno = "_".join(phenodb_id[2:])
                        unique_id = "{0}_{1}".format(phenodb_id[0], accno)

                    self.id_to_idwithvar[unique_id] = id_with_group

                    # ab_class = self.get_csv_tuple(line_list[1].lower())

                    pub_phenotype = self.get_csv_tuple(line_list[2].lower())
                    if("unknown" in pub_phenotype or "none" in pub_phenotype):
                        pub_phenotype = ()

                    abs = []
                    for ab_name in pub_phenotype:
                        # TODO: Fix database
                        if(ab_name == "see notes"):
                            continue
                        classes = self.dis_class_defs.get(ab_name, None)
                        if(classes is None):
                            eprint("Entry {id:s} contained antibiotic "
                                   "'{ab:s}' not found in ab class "
                                   "def file."
                                   .format(id=unique_id, ab=ab_name))
                            continue
                        ab = Antibiotics(name=ab_name, classes=classes)
                        abs.append(ab)
                        for class_ in classes:
                            if(class_ in self.antibiotics):
                                self.antibiotics[class_][ab] = True
                            else:
                                self.antibiotics[class_] = {}
                                self.antibiotics[class_][ab] = True

                    # phenotype = list(pub_phenotype)

                    pmid = self.get_csv_tuple(line_list[3].lower())

                    if(len(line_list) > 4 and line_list[4]):
                        res_mechanics = line_list[4].lower()
                    else:
                        res_mechanics = ""

                    if(len(line_list) > 5 and line_list[5]):
                        notes = line_list[5]
                    else:
                        notes = ""

                    # Line list pos 6 is required genes

                    #
                    # The following should be written out
                    #
                    # if(len(line_list) > 6 and line_list[6]):
                    #    sug_phenotype = self.get_csv_tuple(line_list[6])
                    #    for p in sug_phenotype:
                    #        if p not in pub_phenotype:
                    #            phenotype.append(p)
                    # else:
                    sug_phenotype = ()

                    if(len(line_list) > 7 and line_list[7]):
                        gene_class = line_list[7].lower()
                    else:
                        gene_class = None
                    if(len(line_list) > 8 and line_list[8]):
                        susceptibile = self.get_csv_tuple(line_list[6])
                    else:
                        susceptibile = ()
                    if(len(line_list) > 9 and line_list[9]):
                        species = self.get_csv_tuple(line_list[9])
                    else:
                        species = None

                    pheno = Phenotype(unique_id, abs,
                                      sug_phenotype, pub_phenotype, pmid,
                                      susceptibile=susceptibile,
                                      gene_class=gene_class, notes=notes,
                                      res_db="DisinFinder")

                    pheno_lst = self.get(unique_id, [])
                    pheno_lst.append(pheno)
                    self[unique_id] = pheno_lst

                except IndexError:
                    eprint("PhenoDB load_acquired_dbacq IndexError in line {0}"
                           .format(line_counter))
                    eprint("Split line:\n{0}".format(str(line_list)))

    @staticmethod
    def if_promoter_rename(s):
        out_string = s

        regex = r"^(.+-promoter)-size-\d+bp([_;]{1,2}\d+[_;]{1,2}.+)$"
        promoter_match = re.search(regex, s)

        if promoter_match:
            out_string = f"{promoter_match.group(1)}{promoter_match.group(2)}"

        return out_string

    @staticmethod
    def if_promoter_gene_rename(s):
        """ 
            Needed instead of if_promoter_rename() when only gene name is
            present without variant number and acc no 
        """
        out_string = s

        regex = r"^(.+-promoter)-size-\d+bp$"
        promoter_match = re.search(regex, s)

        if promoter_match:
            out_string = f"{promoter_match.group(1)}"

        return out_string

    @staticmethod
    def get_csv_tuple(csv_string, sep=",", lower=True):
        """ Takes a string containing a comma seperated list, makes it all
            lower case, remove empty entries, remove trailing and preseeding
            whitespaces, and returns the list as a tuple.
        """
        if(lower):
            csv_string = csv_string.lower()

        string_list = csv_string.split(sep)
        # Remove empty entries.
        out_list = [var.strip() for var in string_list if var]
        return tuple(out_list)

    def print_db_stats(self):
        """ Prints some stats about the database to stdout.
        """
        counter_ab_class = 0
        counter_ab = 0

        ab_output = ""

        print("-------------- LIST OF CLASSES --------------")
        for ab_class in self.antibiotics:
            counter_ab_class += 1
            print(ab_class)
        print("--------------- END OF CLASSES --------------")
        print("|")
        print("|")
        print("------------ LIST OF ANTIBIOTICS ------------")
        for ab_class in self.antibiotics:
            print(ab_class)
            for ab in self.antibiotics[ab_class]:
                counter_ab += 1
                print("\t" + ab)
        print("------------ END OF ANTIBIOTICS -------------")
        print("|")
        print("|")
        print("------------------ SUMMARY ------------------")
        print("No. of classes: " + str(counter_ab_class))
        print("No. of antibiotics: " + str(counter_ab))
        print("-------------------- END --------------------")


class MutationGenotype(object):
    """
    """
    def __init__(self, mut_string):
        mut_match = re.search(r"^(.+)_(\D+)(\d+)(.+)$", mut_string)
        self.gene = mut_match.group(1).replace("_", ";;")
        self.ref = mut_match.group(2).lower()
        self.pos = mut_match.group(3)
        alt_str = mut_match.group(4).lower()
        self.alternatives = tuple(alt_str.split("."))
        self.mut_id_prefix = "".join((self.gene, "_", self.ref, self.pos))

    def create_mut(self, mut_aa):
        """ Given an aa alternative, creates a Mutation object.
        """
        unique_id = self.mut_id_prefix + "_" + mut_aa
        mut = Mutation(unique_id, seq_region=gene, pos=self.pos,
                       ref_aa=self.ref, mut_aa=mut_aa)
        # Check if the created mutation is part of the MutationGenotype
        if(self.is_in([mut])):
            return mut
        else:
            sys.exit("An attempt was made to create a mutation from a"
                     "MutationGenotype that was not defined in that "
                     "genotype.")

    def is_in(self, feat_dict):
        """ Checks if any of the mutation genotypes exists in the given
            feature dict. Retuns feature if found. Returns False
            otherwise.
        """
        for feat_id in feat_dict:
            feats = feat_dict[feat_id]
            feat = feats[0]
            if(self == feats[0]):
                return feat

        return False

    def __eq__(self, other):
        """ Compare MutationGenotype to other MutationGenotypes,
            Mutation objects, or strings.

            Returns the matched mutation aa if compared to a mutation
            or a string. Returns boolean if compared to another
            MutationGenotype object.
        """
        if isinstance(other, Mutation):
            other_prefix = f"{other.seq_region}_{other.ref_aa}{other.pos}"
            if self.mut_id_prefix == other_prefix:
                if other.mut_aa in self.alternatives:
                    return other.mut_aa
            return False

        elif isinstance(other, MutationGenotype):

            # Returns True if gene, ref base, position and the tuple of
            # alternatives match.
            if(self.mut_id_prefix == other.mut_id_prefix):

                if(len(self.alternatives) != len(other.alternatives)):
                    return False
                for self_alt in self.alternatives:
                    if(self_alt not in other.alternatives):
                        return False
                return True
            else:
                return False

        elif isinstance(other, str):
            raise NotImplementedError("Attempted to compare MutationGenotype "
                                      "to str.")

    def __ne__(self, other):
        result = self.__eq__(other)
        if result is NotImplemented:
            return result
        return not result


class Phenotype(object):
    """ A Phenotype object describes the antibiotics a feature/gene causes
        resistance and susceptibility against.
        unique_id: the id is used to locate the specified phenotype.
        phenotype: Tuple of antibiotics that gene causes resistance toward.
        ab_class: Class of resistance (e.g. Beta-Lactamase).
        pub_phenotype: Tuple of published resistance (antibiotics).
        pmid: Tuple of PubMed IDs of publications presenting resistance and
              susceptibility.
        susceptibility: Tuple of antibiotics that the gene is known to be
                        susceptibil towards.
        gene_class: resistance gene class (e.g. class D)
        notes: String containing other information on the resistance gene.
        species: Ignore resistance in these species as it is intrinsic.
        req_mut: Mutations required to cause the penotype. This is a tuple of
                 tuples of MutationGenotypes. The outmost tuple represent
                 groups of mutations that are needed for the phenotype,
                 but each of the groups are independent of each other. The
                 inner tuples represent individual mutations for a specific
                 reference aa/nucs found at the position. The MutationGenotype
                 objects contain information on known mutations.
    """
    def __init__(self, unique_id, antibiotics, sug_phenotype,
                 pub_phenotype, pmid, susceptibile=(), gene_class=None,
                 notes="", species=None, res_mechanics=None, req_muts=None,
                 res_db=None):
        self.unique_id = unique_id
        self.antibiotics = antibiotics

        self.sug_phenotype = sug_phenotype
        self.pub_phenotype = pub_phenotype
        self.pmid = pmid
        self.susceptibile = susceptibile
        self.gene_class = gene_class
        self.notes = notes
        self.res_mechanics = res_mechanics
        self.req_muts = req_muts
        self.res_database = res_db

class Antibiotics(object):
    """ Class is implemented to be key in a dict. The class can be tested
        against isinstances of itself and strings.
    """
    def __init__(self, name, classes, feature=None, published=None):
        self.name = name
        self.classes = classes
        self.published = published
        self.features = {}
        if(feature):
            self.add_feature(feature)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, Antibiotics):
            return other.name == self.name
        elif isinstance(other, str):
            return other == self.name
        else:
            return NotImplemented

    def __ne__(self, other):
        result = self.__eq__(other)
        if result is NotImplemented:
            return result
        return not result

    def add_feature(self, feature):
        feat_list = self.features.get(feature.unique_id, [])
        feat_list.append(feature)
        self.features[feature.unique_id] = feat_list

    def get_mut_names(self, _list=False):
        names = {}
        for f in self.features:
            feature = self.features[f]
            if(not (isinstance(feature, Mutation)
                    or isinstance(feature, FeatureGroup))):
                continue

            if(isinstance(feature, FeatureGroup)):
                feature_lst = list(feature.values())
            else:
                feature_lst = [feature]

            for entry in feature_lst:
                names[entry.unique_id] = [entry.seq_region,
                                          entry.mut_string_short]

        if(_list):
            return names.keys()
        else:
            return names

    def get_mut_namewannot(self, tostring=False):
        names = self.get_mut_names()

        output = []
        for unique_id in names:
            annot = names[unique_id]
            name_w_annot = annot[0] + " (" + "".join(annot[1:]) + ")"
            output.append(name_w_annot)

        if(tostring):
            return ", ".join(output)
        else:
            return output

    def get_gene_names(self, list_=False):
        names = {}
        for feat_list in self.features.values():
            for feature in feat_list:
                if(not isinstance(feature, ResGene)):
                    continue

                if(feature.hit and feature.hit.name):
                    names[feature.unique_id] = feature.hit.name
                else:
                    names[feature.unique_id] = "Not Available"

        if(list_):
            return names.keys()
        else:
            return names

    def get_gene_namewacc(self, tostring=False):
        names = self.get_gene_names()

        output = []
        for acc in names:
            name_w_acc = names[acc] + " (" + acc + ")"
            output.append(name_w_acc)

        if(tostring):
            return ", ".join(output)
        else:
            return output

    def get_pubmed_ids(self, phenodb):
        pmids = {}
        for feature in self.features:
            phenotypes = phenodb[feature]
            for phenotype in phenotypes:
                for pmid in phenotype.pmid:
                    pmids[pmid] = True
        return tuple(pmids.keys())


class ResProfile(object):
    """ Given a list of features and a PhenoDB object, an object is created
        that will contain a resistance profile based on the features given and
        the phenotypes described in the PhenoDB object.
        The class contains an add_feature method that makes it possible to add
        features after the object has been created. Per default the add_feature
        method will call the update_profile method, but this can be turned off
        if one needs to add a lot of genes and wish to call the update_profile
        method only once after adding the features.
    """
    def __init__(self, features, phenodb):
        self.phenodb = phenodb
        self.resistance = {}
        self.features = {}
        self.susceptibile = {}
        self.resistance_classes = {}
        self.missing_db_features = []
        self.unknown_db_features = []

        for feature in features:
            if(feature.unique_id in phenodb):
                # Add feature
                if feature.unique_id in self.features:
                    self.features[feature.unique_id].append(feature)
                else:
                    self.features[feature.unique_id] = [feature]
                # Several phenotypes can exist for a single feature ID.
                for phenotype in phenodb[feature.unique_id]:
                    if(phenotype.res_database == feature.ref_db):
                        if(phenotype.antibiotics):
                            self.add_phenotype(feature, phenotype,
                                               update=False)
                        else:
                            self.unknown_db_features.append(feature)
                            self.update_classes_dict_of_feature_sets(
                                self.resistance_classes, feature)

            else:
                self.missing_db_features.append(feature)
                self.update_classes_dict_of_feature_sets(
                    self.resistance_classes, feature)
        self.update_profile()

    @staticmethod
    def update_classes_dict_of_feature_sets(classes, feature):
        """
        """
        if(isinstance(feature, ResGene) or isinstance(feature, ResMutation)):
            for _class in feature.ab_class:
                if(_class not in classes):
                    classes[_class] = set()
                classes[_class].add(feature)

    def add_phenotype(self, feature, phenotype, update=True):
        """
        """
        # Handle required mutations
        if(phenotype.req_muts is not None):

            # Iterate through the different groups of mutations (See
            # Phenotype description).
            mut_found = False
            mut_groups_found = []
            for mut_group in phenotype.req_muts:
                feat_group = FeatureGroup([feature])

                # Iterate through the individual mutations.
                for mut in mut_group:
                    mut_feat = mut.is_in(self.features)
                    if(mut_feat):
                        feat_group[mut_feat.unique_id] = mut_feat
                        mut_found = True
                    # Mutation not found, the requied mutations in this
                    # group are therefore not fullfilled.
                    else:
                        mut_found = False
                        break
                # All mutations from a group was found.
                if(mut_found is True):
                    mut_groups_found.append(feat_group)
                    break
            # No group contained all required mutations.
            if(mut_found is False):
                return
            else:
                # If several mut groups exist, then choose the one that
                # contains most req mutations.
                feature = {}
                for fg in mut_groups_found:
                    if(len(fg) > len(feature)):
                        feature = fg

        for antibiotic in phenotype.antibiotics:
            resprofile_ab = self.resistance.get(antibiotic, None)
            if(resprofile_ab):
                antibiotic = resprofile_ab

            # Create collection of features grouped with respect to ab class
            for _class in antibiotic.classes:
                if(_class not in self.resistance_classes):
                    self.resistance_classes[_class] = set()
                self.resistance_classes[_class].add(feature)
            antibiotic.add_feature(feature)
            self.resistance[antibiotic.name] = antibiotic

        # TODO: delete this for-loop
        for antibiotic in phenotype.sug_phenotype:

            # Create collection of features grouped with respect to ab class
            for _class in phenotype.ab_class:
                if(_class not in self.resistance_classes):
                    self.resistance_classes[_class] = set()
                self.resistance_classes[_class].add(feature)

            if(antibiotic not in self.resistance):
                self.resistance[antibiotic] = Antibiotics(antibiotic,
                                                          phenotype.ab_class,
                                                          feature,
                                                          published=False)
            else:
                self.resistance[antibiotic].add_feature(feature)

        # TODO: delete this for-loop
        for antibiotic in phenotype.susceptibile:
            if(antibiotic not in self.resistance):
                self.susceptibile[antibiotic] = Antibiotics(antibiotic,
                                                            phenotype.ab_class,
                                                            feature)

        if(update):
            self.update_profile()

    def update_profile(self):
        """
        """
        susc_abs = self.susceptibile.keys()
        for ab in susc_abs:
            if ab in self.resistance:
                del self.susceptibile[ab]


class FeatureGroup(dict):
    """ This class is a dict with a unique_id variable created from the
        keys in the dict.

        LIMITATION: The class does not handle the deletion of keys,
                    only additions.
    """
    def __init__(self, features):
        feature_ids = []
        for feature in features:
            feature_ids.append(feature.unique_id)
            super(FeatureGroup, self).__setitem__(feature.unique_id, feature)
            # self[feature.unique_id] = feature

        self.unique_id = "_".join(feature_ids)

    def __hash__(self):
        return hash(self.unique_id)

    def __setitem__(self, k, v):
        self.unique_id = self.unique_id + "_" + k
        return super(FeatureGroup, self).__setitem__(k, v)



#!/usr/bin/env python3

import os.path
import os
import sys
import subprocess

from cgelib.utils.loaders_mixin import LoadersMixin

from .pointfinder import PointFinder


class Config():

    ENV_VAR_FILENAME = "environment_variables.md"
    SPECIES_ABBR_FILENAME = "species_abbreviations.md"

    DEFAULT_VALS = {
        "inputfasta": None,
        "inputfastq": None,
        "nanopore": False,
        "outputPath": None,
        "out_json": None,
        "blastPath": "blastn",
        "kmaPath": "kma",
        "species": None,
        "ignore_missing_species": False,
        "db_path_res": None,
        "db_path_res_kma": None,
        "acquired": None,
        "acq_overlap": 30,
        "min_cov": 0.6,
        "threshold": 0.8,
        "point": None,
        "db_path_point": None,
        "db_path_point_kma": None,
        "specific_gene": None,
        "unknown_mut": False,
        "min_cov_point": 0.01,
        "threshold_point": 0.8,
        "ignore_indels": False,
        "ignore_stop_codons": False,
        "output_aln": False,
        "pickle": False
    }

    def __init__(self, args):

        # Directoy of config.py substracted the last dir 'cge'
        self.resfinder_root = os.path.dirname(os.path.realpath(__file__))[:-3]
        self.env_var_file = "{}{}".format(self.resfinder_root,
                                          Config.ENV_VAR_FILENAME)
        self.species_abbr_file = "{}{}".format(self.resfinder_root,
                                               Config.SPECIES_ABBR_FILENAME)
        Config.set_default_and_env_vals(args, self.env_var_file)

        self.set_general_opts(args)
        if(self.acquired):
            self.set_resfinder_opts(args)
        if(self.point):
            self.set_pointfinder_opts(args)
        if(self.disinf):
            self.set_disinfinder_opts(args)
        self.set_phenotype_opts(args)

        if(self.acquired is False and self.point is False
           and self.disinf is False):
            sys.exit("Please specify to look for acquired resistance genes, "
                     "chromosomal mutaitons or both!\n")

    def set_general_opts(self, args):
        self.outputPath = os.path.abspath(args.outputPath)
        os.makedirs(self.outputPath, exist_ok=True)

        if(args.out_json):
            if not args.out_json.endswith(".json"):
                sys.exit("Please specify the path to the JSON file including "
                         "its filename ending with .json.\n")
            self.out_json = os.path.abspath(args.out_json)
            os.makedirs(os.path.dirname(self.out_json), exist_ok=True)
        else:
            self.out_json = False

        self.acquired = bool(args.acquired)
        self.point = bool(args.point)
        self.disinf = bool(args.disinfectant)
        self.species = Config.get_species(args.species, self.species_abbr_file)

        if(args.inputfasta):
            self.set_fasta_related_opts(args)

        if(args.inputfastq):
            self.set_fastq_related_opts(args)

        amr_abbreviations_file = ("{}/amr_abbreviations.md"
                                  .format(self.resfinder_root))
        self.amr_abbreviations = LoadersMixin.load_md_table_after_keyword(
            amr_abbreviations_file, "## Abbreviations")
        self.output_aln = bool(args.output_aln)
        self.pickle = args.pickle

    @staticmethod
    def get_species(in_species, species_def_filepath):
        out_species = in_species
        if(in_species is not None and in_species.lower() == "other"):
            out_species = "other"
        elif(in_species is not None):
            out_species = in_species.lower()

        species_transl = LoadersMixin.load_md_table_after_keyword(
            species_def_filepath, "## Species Abbreviations Table",
            header_key="Abbreviation")

        fixed_species = species_transl.get(out_species, None)
        if(fixed_species):
            out_species = fixed_species[0]

        return out_species

    def set_fasta_related_opts(self, args):
        self.inputfastq_1 = None
        self.inputfasta = self.get_abs_path_and_check(args.inputfasta)

        self.outPath_res_blast = "{}/resfinder_blast".format(self.outputPath)
        os.makedirs(self.outPath_res_blast, exist_ok=True)

        self.outPath_disinf_blast = ("{}/disinfinder_blast"
                                     .format(self.outputPath))
        os.makedirs(self.outPath_disinf_blast, exist_ok=True)

        self.outPath_point_blast = ("{}/pointfinder_blast"
                                    .format(self.outputPath))
        os.makedirs(self.outPath_point_blast, exist_ok=True)

        self.sample_name = os.path.basename(self.inputfasta)
        self.method = PointFinder.TYPE_BLAST
        self.blast = self.get_prg_path(args.blastPath)
        self.kma = None

    def set_fastq_related_opts(self, args):
        self.inputfasta = None
        self.inputfastq_1 = self.get_abs_path_and_check(
            args.inputfastq[0])
        if(len(args.inputfastq) == 2):
            self.inputfastq_2 = self.get_abs_path_and_check(
                args.inputfastq[1])
        elif(len(args.inputfastq) > 2):
            sys.exit("ERROR: More than 2 files were provided to inputfastq: "
                     "{}.".format(args.inputfastq))
        else:
            self.inputfastq_2 = None

        self.outPath_res_kma = "{}/resfinder_kma".format(self.outputPath)
        os.makedirs(self.outPath_res_kma, exist_ok=True)

        self.outPath_disinf_kma = "{}/disinfinder_kma".format(self.outputPath)
        os.makedirs(self.outPath_disinf_kma, exist_ok=True)

        self.outPath_point_kma = "{}/pointfinder_kma".format(self.outputPath)
        os.makedirs(self.outPath_point_kma, exist_ok=True)

        self.sample_name = os.path.basename(args.inputfastq[0])
        self.method = PointFinder.TYPE_KMA
        self.kma = self.get_prg_path(args.kmaPath)
        self.nanopore = args.nanopore

    def set_resfinder_opts(self, args):
        self.set_path_resfinderdb(args)
        self.db_config_file = f"{self.db_path_res}/config"
        self.db_notes_file = f"{self.db_path_res}/notes.txt"
        self.db_panels_file = f"{self.db_path_res}/phenotype_panels.txt"
        if not os.path.exists(self.db_config_file):
            sys.exit("Input Error: The database config file could not be found"
                     " in the ResFinder database directory.")
        if not os.path.exists(self.db_notes_file):
            sys.exit("Input Error: The database notes.txt file could not be "
                     "found in the ResFinder database directory.")
        if not os.path.exists(self.db_panels_file):
            sys.exit("Input Error: The database phenotype_panels.txt file "
                     "could not be found in the ResFinder database directory.")

        args.min_cov = float(args.min_cov)
        args.threshold = float(args.threshold)

        # Check if coverage/identity parameters are valid
        if(args.min_cov > 1.0 or args.min_cov < 0.0):
            sys.exit("ERROR: Minimum coverage above 1 or below 0 is not "
                     "allowed. Please select a minimum coverage within the "
                     "range 0-1 with the flag -l. Given value: {}."
                     .format(args.min_cov))
        self.rf_gene_cov = args.min_cov

        if(args.threshold > 1.0 or args.threshold < 0.0):
            sys.exit("ERROR: Threshold for identity of ResFinder above 1 or "
                     "below 0 is not allowed. Please select a threshold for "
                     "identity within the range 0-1 with the flag -t. Given "
                     "value: {}.".format(args.threshold))
        self.rf_gene_id = args.threshold

        self.rf_overlap = int(args.acq_overlap)

    def set_disinfinder_opts(self, args):
        self.set_path_disinfinderdb(args)
        self.db_config_disinf_file = "{}/config".format(self.db_path_disinf)
        self.db_notes_disinf_file = "{}/notes.txt".format(self.db_path_disinf)
        if not os.path.exists(self.db_config_disinf_file):
            sys.exit("Input Error: The database config file could not be found"
                     f" at: {self.db_config_disinf_file}")
        if not os.path.exists(self.db_notes_disinf_file):
            sys.exit("Input Error: The database notes.txt file could not be "
                     f"found at: {self.db_notes_disinf_file}")

        args.min_cov = float(args.min_cov)
        args.threshold = float(args.threshold)

        # Check if coverage/identity parameters are valid
        if(args.min_cov > 1.0 or args.min_cov < 0.0):
            sys.exit("ERROR: Minimum coverage above 1 or below 0 is not "
                     "allowed. Please select a minimum coverage within the "
                     "range 0-1 with the flag -l. Given value: {}."
                     .format(args.min_cov))
        self.dis_gene_cov = args.min_cov

        self.dis_overlap = int(args.acq_overlap)

        if(args.threshold > 1.0 or args.threshold < 0.0):
            sys.exit("ERROR: Threshold for identity of DisinFinder above 1 or "
                     "below 0 is not allowed. Please select a threshold for "
                     "identity within the range 0-1 with the flag -t. Given "
                     "value: {}.".format(args.threshold))
        self.dis_gene_id = args.threshold

    def set_pointfinder_opts(self, args):
        if(not self.species and not args.ignore_missing_species):
            sys.exit("ERROR: Chromosomal point mutations cannot be located if "
                     "no species has been provided. Please provide species "
                     "using the --species option.")
        elif((not self.species and args.ignore_missing_species)
                or self.species.lower() == 'other'):
            self.point = False
            return

        self.set_path_pointdb(args)
        if(self.db_path_point is None):
            self.point = False
            return

        self.specific_gene = args.specific_gene

        args.min_cov_point = float(args.min_cov_point)
        args.threshold_point = float(args.threshold_point)

        # Check if coverage/identity parameters are valid
        if(args.min_cov_point > 1.0 or args.min_cov_point < 0.0):
            sys.exit("ERROR: Minimum coverage above 1 or below 0 is not "
                     "allowed. Please select a minimum coverage within the "
                     "range 0-1 with the flag -l. Given value: {}."
                     .format(args.min_cov_point))
        self.pf_gene_cov = args.min_cov_point

        if(args.threshold_point > 1.0 or args.threshold_point < 0.0):
            sys.exit("ERROR: Threshold for identity of ResFinder above 1 or "
                     "below 0 is not allowed. Please select a threshold for "
                     "identity within the range 0-1 with the flag -t. Given "
                     "value: {}.".format(args.threshold_point))
        self.pf_gene_id = args.threshold_point

        self.unknown_mut = args.unknown_mut
        self.ignore_indels = args.ignore_indels
        self.ignore_stop_codons = args.ignore_stop_codons

    def set_phenotype_opts(self, args):
        self.point_file = None
        if(self.point):
            if os.path.exists("{}/phenotypes.txt"
                              .format(self.db_path_point)):
                self.point_file = ("{}/phenotypes.txt"
                                   .format(self.db_path_point))
                _ = self.get_abs_path_and_check(self.point_file)
            elif os.path.exists("{}/resistens-overview.txt"
                                .format(self.db_path_point)):
                self.point_file = ("{}/resistens-overview.txt"
                                   .format(self.db_path_point))
                _ = self.get_abs_path_and_check(self.point_file)
            else:
                sys.exit("Error: The pointfinder database does not have the "
                         "'phenotypes.txt' file (new database) neither the "
                         "'resistens-overview.txt' file (old database)")
        if(not args.acquired):
            self.set_path_resfinderdb(args)

        self.db_panels_file = f"{self.db_path_res}/phenotype_panels.txt"
        _ = self.get_abs_path_and_check(self.db_panels_file)

        self.abclassdef_file = f"{self.db_path_res}/antibiotic_classes.txt"
        _ = self.get_abs_path_and_check(self.abclassdef_file)

        if(self.point or self.acquired):
            self.phenotype_file = ("{}/phenotypes.txt"
                                   .format(self.db_path_res))
            _ = self.get_abs_path_and_check(self.phenotype_file)
        else:
            self.phenotype_file = None

        if(self.disinf):
            self.disinf_file = ("{}/phenotypes.txt".format(self.db_path_disinf)
                                )
            _ = self.get_abs_path_and_check(self.disinf_file)

            self.disclassdef_file = ("{}/disinfectant_classes.txt"
                                     .format(self.db_path_disinf))
            _ = self.get_abs_path_and_check(self.disclassdef_file)
        else:
            self.disinf_file = None
            self.disclassdef_file = None

    @staticmethod
    def get_abs_path_and_check(path, allow_exit=True):
        abs_path = os.path.abspath(path)
        if(not os.path.isfile(abs_path) and not os.path.isdir(abs_path)):
            if(allow_exit):
                sys.exit("ERROR: Path not found: {}".format(path))
            else:
                raise FileNotFoundError
        return abs_path

    def set_path_pointdb(self, args):
        tmp_list = self.species.split()
        if(len(tmp_list) != 1 and len(tmp_list) != 2):
            sys.exit("ERROR: Species name must contain 1 or 2 names. Given "
                     "value: {}".format(self.species))

        if(len(tmp_list) == 2):
            tmp_species_dir = "_".join(tmp_list)
        else:
            tmp_species_dir = tmp_list[0]

        if(args.db_path_point is None):
            args.db_path_point = "{}{}".format(self.resfinder_root,
                                               "/db_pointfinder")

        path_pointdb = os.path.abspath(args.db_path_point)

        self.species_dir = self._parse_species_dir(path_pointdb,
                                                   tmp_species_dir,
                                                   args.ignore_missing_species)

        if(self.species_dir is not None):
            self.db_path_point_root = path_pointdb
            self.db_path_point = "{}/{}".format(path_pointdb, self.species_dir)

    def set_path_disinfinderdb(self, args):
        self.db_path_disinf = args.db_path_disinf
        if(self.db_path_disinf is None):
            self.db_path_disinf = "{}{}".format(self.resfinder_root,
                                                "/db_disinfinder")
        self.db_path_disinf_kma = args.db_path_disinf_kma
        if(self.db_path_disinf_kma is None):
            self.db_path_disinf_kma = self.db_path_disinf

        try:
            self.db_path_disinf = Config.get_abs_path_and_check(
                self.db_path_disinf, allow_exit=False)
        except FileNotFoundError:
            sys.exit("Could not locate DisinFinder database path: {}"
                     .format(self.db_path_disinf))

        try:
            self.db_path_disinf_kma = Config.get_abs_path_and_check(
                self.db_path_disinf_kma, allow_exit=False)
        except FileNotFoundError:
            if(self.disinf and self.inputfastq_1):
                sys.exit("Could not locate DisinFinder database index path: {}"
                         .format(self.db_path_disinf_kma))
            else:
                pass

    def set_path_resfinderdb(self, args):
        self.db_path_res = args.db_path_res
        if(self.db_path_res is None):
            self.db_path_res = "{}{}".format(self.resfinder_root,
                                             "/db_resfinder")
        self.db_path_res_kma = args.db_path_res_kma
        if(self.db_path_res_kma is None):
            self.db_path_res_kma = self.db_path_res

        try:
            self.db_path_res = Config.get_abs_path_and_check(
                self.db_path_res, allow_exit=False)
        except FileNotFoundError:
            if not self.acquired:
                sys.exit("ResFinder database is needed even if only searching "
                         "for point mutations or disinfectant genes. Could not"
                         " locate ResFinder database path: "
                         f"{self.db_path_res}")
            else:
                sys.exit("Could not locate ResFinder database path: "
                         f"{self.db_path_res}")

        try:
            self.db_path_res_kma = Config.get_abs_path_and_check(
                self.db_path_res_kma, allow_exit=False)
        except FileNotFoundError:
            if(self.acquired and self.inputfastq_1):
                sys.exit("Could not locate ResFinder database index path: {}"
                         .format(self.db_path_res_kma))
            else:
                pass

    def _parse_species_dir(self, path_pointdb, species_dir,
                           ignore_missing_species):
        # Check if a database for species exists
        point_dbs = PointFinder.get_db_names(path_pointdb)
        if(species_dir not in point_dbs):
            # If no db for species is found check if db for genus is found
            # and use that instead
            tmp_list = self.species.split()
            if(tmp_list[0] in point_dbs):
                species_dir = tmp_list[0]
            elif(ignore_missing_species):
                self.species = None
                self.db_path_point_root = None
                self.db_path_point = None
                species_dir = None
            else:
                sys.exit("ERROR: species '{}' ({}) does not seem to exist"
                         " as a PointFinder database."
                         .format(self.species, species_dir))
        return species_dir

    @staticmethod
    def get_prg_path(prg_path):
        try:
            prg_path = Config.get_abs_path_and_check(prg_path,
                                                     allow_exit=False)
        except FileNotFoundError:
            pass

        try:
            _ = subprocess.check_output([prg_path, "-h"])
        except PermissionError:
            sys.exit("ERROR: Missing permission. Unable to execute app from"
                     " the path: {}".format(prg_path))
        return prg_path

    @staticmethod
    def set_default_and_env_vals(args, env_def_filepath):

        known_envs = LoadersMixin.load_md_table_after_keyword(
            env_def_filepath, "## Environment Variables Table")

        # Set flag values defined in environment variables
        for var, entries in known_envs.items():

            try:
                cli_val = getattr(args, entries[0])
                # Flags set by user will not be None, default vals will be None
                if(cli_val is not None):
                    continue

                var_val = os.environ.get(var, None)
                if(var_val is not None):
                    setattr(args, entries[0], var_val)

            except AttributeError:
                sys.exit("ERROR: A flag set in the Environment Variables Table"
                         " in the README file did not match any valid flags in"
                         " ResFinder. Flag not recognized: {}."
                         .format(entries[0]))

        Config._set_default_values(args)

    @staticmethod
    def _set_default_values(args):
        for flag, def_val in Config.DEFAULT_VALS.items():
            val = getattr(args, flag)
            if(val is None):
                setattr(args, flag, def_val)



#!/usr/bin/env python3


class Phenotype():
    """ A Phenotype object describes the antibiotics a feature/gene causes
        resistance and susceptibility against.
        unique_id: the id is used to locate the specified phenotype.
        phenotype: Tuple of antibiotics that gene causes resistance toward.
        ab_class: Class of resistance (e.g. Beta-Lactamase).
        pub_phenotype: Tuple of published resistance (antibiotics).
        pmid: Tuple of PubMed IDs of publications presenting resistance and
              susceptibility.
        susceptibility: Tuple of antibiotics that the gene is known to be
                        susceptibil towards.
        gene_class: resistance gene class (e.g. class D)
        notes: String containing other information on the resistance gene.
        species: Species exceptions, where resistance is not observed. NOTE:
                 This information is not yet used for anything.
    """
    def __init__(self, unique_id, phenotype, ab_class, sug_phenotype,
                 pub_phenotype, pmid, susceptibile=(), gene_class=None,
                 notes="", species=None, res_mechanics=None):
        self.unique_id = unique_id
        self.phenotype = phenotype
        self.ab_class = ab_class

        self.sug_phenotype = sug_phenotype
        self.pub_phenotype = pub_phenotype
        self.pmid = pmid
        self.susceptibile = susceptibile
        self.gene_class = gene_class
        self.notes = notes
        self.species = species



#!/usr/bin/env python3
import random
import string

from ..phenotype2genotype.res_profile import PhenoDB


class GeneResult(dict):
    def __init__(self, res_collection, res, db_name, conf=None, alignments=None):
        """
            Input:
                res_collection: Result object created by the cgelib package.
                res: Custom dictionary containing information about a single hit
                     from ResFinder.
                db_name: 'ResFinder' or 'PointFinder' or 'DisinFinder'

            Method creates a seq_region dict as defined in the BeOne template:
            https://bitbucket.org/genomicepidemiology/cgelib/src/master/src/
            cgelib/output/templates_json/beone/
            from res.
        """
        self.db_name = db_name
        self["type"] = "seq_region"
        self["gene"] = True

        self["ref_id"] = res["sbjct_header"]
        self["ref_id"] = PhenoDB.if_promoter_rename(self["ref_id"])
        self["name"], self.variant, self["ref_acc"] = (
            GeneResult._split_sbjct_header(self["ref_id"]))
        self["ref_database"] = [res_collection.get_db_key(db_name)[0]]
        self["identity"] = res["perc_ident"]
        self["alignment_length"] = res["HSP_length"]
        self["ref_seq_length"] = res["sbjct_length"]
        self["depth"] = res.get("depth", None)
        self["ref_start_pos"] = res["sbjct_start"]
        self["ref_end_pos"] = res["sbjct_end"]
        self["query_id"] = res["contig_name"]    # Positional essential
        self["query_start_pos"] = res["query_start"]    # Positional essential
        self["query_end_pos"] = res["query_end"]    # Positional essential
        self["pmids"] = []
        self["notes"] = []

        if conf and conf.output_aln:
            hit_class, hit_key = self.get_hit_class(alignments.gene_align_query)

            self["query_string"] = alignments.gene_align_query[hit_class][
                hit_key]
            self["alignment_string"] = alignments.gene_align_homo[hit_class][
                hit_key]
            self["ref_string"] = alignments.gene_align_sbjct[hit_class][hit_key]

        # BLAST coverage formatted results
        coverage = res.get("coverage", None)
        if (coverage is None):
            # KMA coverage formatted results
            coverage = res["perc_coverage"]
        else:
            coverage = float(coverage) * 100
        self["coverage"] = coverage

        self["grade"] = GeneResult.calc_gene_grade(
            coverage=self["coverage"], identity=self["identity"])

        self.remove_NAs()
        uniqueness = self._get_unique_gene_key(res_collection)
        self["key"] = uniqueness

    def get_hit_class(self, gene_align_query):
        """
            Input:
                gene_align_query: Dict from either FinderResult (KMA) or
                                  Blaster (BLAST)
            Output:
                hit_class: Name of AB class
                hit_key: Key used for query (hit)
            
            Method looks through all the queries that align to a reference and
            finds the query (hit) that matches the GeneResult object.
            For BLAST hits the query is named after the coresponding contig
            header (query_id), but for KMA the query key is named after the
            reference it aligns to (ref_id).
        """
        lookup_key = self["query_id"]
        if self["depth"] is not None:
            lookup_key = self["ref_id"]

        for ab_class, hits in gene_align_query.items():            
            if hits:
                hits_keys = list(hits.keys())
                hit_key = next((key for key in hits_keys
                                if key.startswith(lookup_key)), None)
                if hit_key is None:
                    continue
                hit_class = ab_class
                break
            else:
                continue
        
        return hit_class, hit_key

    @staticmethod
    def calc_gene_grade(coverage: float, identity: float) -> int:
        if coverage == 100.0 and identity == 100.0:
            return 3
        elif coverage == 100.0:
            return 2
        else:
            return 1

    def remove_NAs(self):
        """
            Remove all entries containing NA og None as values.

            Removing None is not necessary as the Result object will ignore all
            entries with None values.
        """
        na_keys = []
        for key, val in self.items():
            if (val == "NA" or val is None):
                na_keys.append(key)
        for key in na_keys:
            del self[key]

    @staticmethod
    def get_rnd_unique_gene_key(gene_key, res_collection,
                                minimum_gene_key, delimiter):
        """
            Input:
                gene_key: None-unique key
                res_collection: Result object created by the cgelib package.
                minimum_key: Key prefix
                delimiter: String used as delimiter inside the returned key.
            Output:
                gene_key: Unique key (string)

            If gene_key is found in res_collection. Creates a unique key by
            appending a random string ton minimum_gene_key.
        """
        while (gene_key in res_collection["seq_regions"]):
            rnd_str = GeneResult.random_string(str_len=4)
            gene_key = ("{key}{deli}{rnd}"
                        .format(key=minimum_gene_key, deli=delimiter,
                                rnd=rnd_str))
        return gene_key

    @staticmethod
    def random_string(str_len=4):
        """
            Output:
                random string of length 'str_len'

            Return a random string of the provided length. The string will only
            consist of lowercase ascii letters.
        """
        letters = string.ascii_lowercase
        return ''.join(random.choice(letters) for i in range(str_len))

    @staticmethod
    def _split_sbjct_header(header):
        """
            Input:
                header: database entry header (ref_header/subject_header)
            Output:
                template: name of entry (string)
                variant: Variant interger (string) or None
                acc: Accession number given by sequnce database (string) or None

            Splits the input header by underscores and returns first list item
            as template. If list is > 1 then variant and acc will also return
            strings. If not they return None.
        """
        sbjct = header.split("_")
        template = sbjct[0]

        if (len(sbjct) > 1):
            variant = sbjct[1]
            acc = "_".join(sbjct[2:])
        else:
            variant = None
            acc = None

        return (template, variant, acc)

    def _get_unique_gene_key(self, res_collection, delimiter=";;"):
        """
            Input:
                res_collection: Result object created by the cgelib package.
                delimiter: String used as delimiter inside the returned key.
            Output:
                key: Unique key for hit

            Creates a unique key for GeneResult instance. Key format depends on
            database. If gene result is considered indentical to an existing
            gene result in the provided res_collection, it will not create a new
            key. Two restults are considered identical if they have the same
            query_id, query_start_pos and query_end_pos. If it is mapping
            results the query_* doesn't exist, and results will never be
            considered identical.
        """
        if "ref_acc" not in self:
            gene_key = ("{name}".format(**self))
        else:
            gene_key = ("{name}{deli}{var}{deli}{ref_acc}".format(
                        deli=delimiter, var=self.variant, **self))
        # Attach random string if key already exists
        minimum_gene_key = gene_key
        if gene_key in res_collection["seq_regions"]:

            query_id = self.get("query_id", "NA")

            # Query id == "NA" when FASTQ
            if (query_id == "NA"):
                res_collection["seq_regions"][gene_key]["ref_database"].extend(
                    self["ref_database"])
                gene_key = None
            # Query id != "NA" when FASTA
            elif (self["query_id"]
                    != res_collection["seq_regions"][gene_key]["query_id"]
                  or self["query_start_pos"]
                    != res_collection["seq_regions"][gene_key]["query_start_pos"]
                  or self["query_end_pos"]
                    != res_collection["seq_regions"][gene_key]["query_end_pos"]):
                gene_key = GeneResult.get_rnd_unique_gene_key(
                    gene_key, res_collection, minimum_gene_key, delimiter)
            else:
                res_collection["seq_regions"][gene_key]["ref_database"].extend(
                    self["ref_database"])
                gene_key = None

        return gene_key



#!/usr/bin/env python3
import io
import sys
import os
import subprocess
from argparse import ArgumentParser
import pickle
import json
import hashlib

from cgelib.output.result import Result
from cgelib.utils.loaders_mixin import LoadersMixin
from cgelib.utils.pliers_mixin import PliersMixin

from resfinder.cge.config import Config
from resfinder.cge.resfinder import ResFinder
from resfinder.cge.pointfinder import PointFinder
from resfinder.cge.output.std_results import ResFinderResultHandler
from resfinder.cge.output.std_results import PointFinderResultHandler

#  Modules used to create the extended ResFinder output (phenotype output)
from resfinder.cge.phenotype2genotype.isolate import Isolate
from resfinder.cge.phenotype2genotype.res_profile import PhenoDB
from resfinder.cge.phenotype2genotype.res_sumtable import ResSumTable
from resfinder.cge.phenotype2genotype.res_sumtable import PanelNameError

from resfinder import __version__


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def get_software_exec_res(conf: Config) -> dict:
    software_exec_res = {
        "type": "software_exec",
        "command": " ".join(sys.argv),
        "parameters": get_call_parameters(conf)
    }
    software_exec_res["key"] = hashlib.sha1(
        bytes(software_exec_res["command"], 'UTF-8')).hexdigest()
    return software_exec_res


def get_call_parameters(conf: Config) -> dict:
    parameters = vars(conf).copy()
    del (parameters['amr_abbreviations'])
    return parameters


def main():

    # version = read_version("VERSION")
    version = __version__

    ##########################################################################
    # PARSE COMMAND LINE OPTIONS
    ##########################################################################

    parser = ArgumentParser(allow_abbrev=False)

    # General options
    parser.add_argument("-ifa", "--inputfasta",
                        help="Input fasta file.",
                        default=None)
    parser.add_argument("-ifq", "--inputfastq",
                        help=("Input fastq file(s). Assumed to be single-end "
                              "fastq if only one file is provided, and assumed"
                              " to be paired-end data if two files are "
                              "provided."),
                        nargs="+",
                        default=None)
    parser.add_argument("--nanopore",
                        action="store_true",
                        dest="nanopore",
                        help="If nanopore data is used",
                        default=False)
    parser.add_argument("-o", "--outputPath",
                        help=("Output directory. If it doesn't exist, it will "
                              "be created."),
                        required=True,
                        default=None)
    parser.add_argument("-j", "--out_json",
                        help=("Specify JSON filename and output directory. If "
                              "the directory doesn't exist, it will be "
                              "created."),
                        default=None)
    parser.add_argument("-b", "--blastPath",
                        help="Path to blastn",
                        default=None)
    parser.add_argument("-k", "--kmaPath",
                        help="Path to KMA",
                        default=None)
    parser.add_argument("-s", "--species",
                        help="Species in the sample",
                        default=None)
    parser.add_argument("--ignore_missing_species",
                        action="store_true",
                        help=("If set, species is provided and --point flag "
                              "is set, will not throw an error if no database "
                              "is found for the provided species. If species "
                              "is not found. Point mutations will silently "
                              "be ignored."),
                        default=False)
    parser.add_argument("--output_aln",
                        action="store_true",
                        help="will add the alignments in the json output.",
                        default=False)

    # Acquired resistance options
    parser.add_argument("-db_res", "--db_path_res",
                        help=("Path to the databases for ResFinder."),
                        default=None)
    parser.add_argument("-db_res_kma", "--db_path_res_kma",
                        help=("Path to the ResFinder databases indexed with "
                              "KMA. Defaults to the value of the --db_res "
                              "flag."),
                        default=None)
    parser.add_argument("-acq", "--acquired",
                        action="store_true",
                        help="Run resfinder for acquired resistance genes",
                        default=None)
    parser.add_argument("-ao", "--acq_overlap",
                        help="Genes are allowed to overlap this number of\
                              nucleotides. Default: {}.".format(
                            Config.DEFAULT_VALS["acq_overlap"]),
                        type=int,
                        default=None)
    parser.add_argument("-l", "--min_cov",
                        help=("Minimum (breadth-of) coverage of ResFinder "
                              "within the range 0-1."),
                        type=float,
                        default=None)
    parser.add_argument("-t", "--threshold",
                        help=("Threshold for identity of ResFinder within the "
                              "range 0-1."),
                        type=float,
                        default=None)
    # Disinfectant resistance options
    parser.add_argument("-d", "--disinfectant",
                        action="store_true",
                        help="Run resfinder for disinfectant resistance genes",
                        default=False)
    parser.add_argument("-db_disinf", "--db_path_disinf",
                        help=("Path to the databases for DisinFinder."),
                        default=None)
    parser.add_argument("-db_disinf_kma", "--db_path_disinf_kma",
                        help=("Path to the DisinFinder databases indexed with "
                              "KMA. Defaults to the value of the --db_res "
                              "flag."),
                        default=None)

    # Point resistance option
    parser.add_argument("-c", "--point",
                        action="store_true",
                        help="Run pointfinder for chromosomal mutations",
                        default=None)
    parser.add_argument("-db_point", "--db_path_point",
                        help="Path to the databases for PointFinder",
                        default=None)
    parser.add_argument("-db_point_kma", "--db_path_point_kma",
                        help=("Path to the PointFinder databases indexed with "
                              "KMA. Defaults to the value of the "
                              "--db_path_point flag."),
                        default=None)
    parser.add_argument("-g", "--specific_gene",
                        nargs='+',
                        help="Specify genes existing in the database to \
                              search for - if none is specified all genes are \
                              included in the search.",
                        default=None)
    parser.add_argument("-u", "--unknown_mut",
                        action="store_true",
                        help=("Show all mutations found even if in unknown to "
                              "the resistance database"),
                        default=None)
    parser.add_argument("-l_p", "--min_cov_point",
                        help=("Minimum (breadth-of) coverage of Pointfinder "
                              "within the range 0-1. If None is selected, the "
                              "minimum coverage of ResFinder will be used."),
                        type=float,
                        default=None)
    parser.add_argument("-t_p", "--threshold_point",
                        help=("Threshold for identity of Pointfinder within "
                              "the range 0-1. If None is selected, the minimum"
                              " coverage of ResFinder will be used."),
                        type=float,
                        default=None)
    parser.add_argument("--ignore_indels",
                        action="store_true",
                        help=("Ignore frameshift-causing indels in "
                              "Pointfinder."),
                        default=None)
    parser.add_argument("--ignore_stop_codons",
                        action="store_true",
                        help="Ignore premature stop codons in Pointfinder.",
                        default=None)
    parser.add_argument("-v", "--version", action="version",
                        version=__version__,
                        help="Show program's version number and exit")

    # Temporary option only available temporary
    parser.add_argument("--pickle",
                        action="store_true",
                        help=("Create a pickle dump of the Isolate object. "
                              "Currently needed in the CGE webserver. "
                              "Dependency and this option is being removed."),
                        default=False)

    args = parser.parse_args()

    # Parse and check all arguments and expected files.
    conf = Config(args)

    # Initialise result dict
    std_result = Result.init_software_result(
        name="ResFinder",
        gitdir=f"{conf.resfinder_root}/../../")

    init_result_data = {
        "provided_species": conf.species,
        "software_version": __version__,
        "key": f"ResFinder-{__version__}",
    }
    std_result.add(**init_result_data)

    if conf.acquired:
        std_result.init_database("ResFinder", conf.db_path_res)
    if conf.point:
        std_result.init_database("PointFinder", conf.db_path_point_root)
    if conf.disinf:
        std_result.init_database("DisinFinder", conf.db_path_disinf)

    std_result.add_class(cl="software_executions", **
                         get_software_exec_res(conf))

    # Load genotype to phenotype database
    res_pheno_db = PhenoDB(
        abclassdef_file=conf.abclassdef_file,
        acquired_file=conf.phenotype_file,
        point_file=conf.point_file,
        disinf_file=conf.disinf_file,
        disclassdef_file=conf.disclassdef_file
    )

    ##########################################################################
    # ResFinder
    ##########################################################################

    if conf.acquired is True:

        blast_results = None
        kma_run = None

        # Actually running ResFinder (for acquired resistance)
        acquired_finder = ResFinder(db_conf_file=conf.db_config_file,
                                    db_path=conf.db_path_res,
                                    pheno_file=conf.phenotype_file,
                                    notes=conf.db_notes_file,
                                    db_path_kma=conf.db_path_res_kma)

        if (conf.inputfasta):
            blast_results = acquired_finder.blast(
                inputfile=conf.inputfasta,
                out_path=conf.outPath_res_blast,
                min_cov=conf.rf_gene_cov,
                threshold=conf.rf_gene_id,
                blast=conf.blast,
                allowed_overlap=conf.rf_overlap
            )

            # DEPRECATED
            # TODO: make a write method that depends on the json output
            acquired_finder.write_results(out_path=conf.outputPath,
                                          result=blast_results,
                                          res_type=ResFinder.TYPE_BLAST,
                                          software="ResFinder")

            ResFinderResultHandler.standardize_results(std_result,
                                                       blast_results,
                                                       "ResFinder",
                                                       conf)

        else:
            kma_nano_args = None
            if conf.nanopore:
                kma_nano_args = '-ont -md 5'

            kma_run = acquired_finder.kma(
                inputfile_1=conf.inputfastq_1,
                inputfile_2=conf.inputfastq_2,
                out_path=conf.outPath_res_kma,
                db_path_kma=conf.db_path_res_kma,
                min_cov=conf.rf_gene_cov,
                threshold=conf.rf_gene_id,
                kma_path=conf.kma,
                databases=acquired_finder.databases,
                sample_name="",
                kma_cge=True,
                kma_apm="p",
                kma_1t1=True,
                kma_add_args=kma_nano_args
            )

            # DEPRECATED
            # TODO: make a write method that depends on the json output
            acquired_finder.write_results(out_path=conf.outputPath,
                                          result=kma_run.results,
                                          res_type=ResFinder.TYPE_KMA,
                                          software="ResFinder")

            ResFinderResultHandler.standardize_results(std_result,
                                                       kma_run,
                                                       "ResFinder",
                                                       conf)
    ##########################################################################
    # DisinFinder
    ##########################################################################
    if (conf.disinf is True):

        blast_results = None
        kma_run = None

        # Actually running DisinFinder (for disinfectant resistance)
        disinf_finder = ResFinder(db_conf_file=conf.db_config_disinf_file,
                                  db_path=conf.db_path_disinf,
                                  pheno_file=conf.disinf_file,
                                  notes=conf.db_notes_disinf_file,
                                  db_path_kma=conf.db_path_disinf_kma)

        if (conf.inputfasta):
            blast_results = disinf_finder.blast(
                inputfile=conf.inputfasta,
                out_path=conf.outPath_disinf_blast,
                min_cov=conf.dis_gene_cov,
                threshold=conf.dis_gene_id,
                blast=conf.blast,
                allowed_overlap=conf.dis_overlap
            )

            # DEPRECATED
            # TODO: make a write method that depends on the json output
            disinf_finder.write_results(out_path=conf.outputPath,
                                        result=blast_results,
                                        res_type=ResFinder.TYPE_BLAST,
                                        software="DisinFinder")

            ResFinderResultHandler.standardize_results(std_result,
                                                       blast_results,
                                                       "DisinFinder",
                                                       conf)

        else:
            kma_nano_args = None
            if conf.nanopore:
                kma_nano_args = '-ont -md 5'

            kma_run = disinf_finder.kma(
                inputfile_1=conf.inputfastq_1,
                inputfile_2=conf.inputfastq_2,
                out_path=conf.outPath_disinf_kma,
                db_path_kma=conf.db_path_disinf_kma,
                min_cov=conf.dis_gene_cov,
                threshold=conf.dis_gene_id,
                kma_path=conf.kma,
                databases=disinf_finder.databases,
                sample_name="",
                kma_cge=True,
                kma_apm="p",
                kma_1t1=True,
                kma_add_args=kma_nano_args
            )

            # DEPRECATED
            # TODO: make a write method that depends on the json output
            disinf_finder.write_results(out_path=conf.outputPath,
                                        result=kma_run.results,
                                        res_type=ResFinder.TYPE_KMA,
                                        software="DisinFinder")

            ResFinderResultHandler.standardize_results(std_result,
                                                       kma_run,
                                                       "DisinFinder",
                                                       conf)
    ##########################################################################
    # PointFinder
    ##########################################################################

    if (conf.point):

        blast_results = None
        kma_run = None

        finder = PointFinder(db_path=conf.db_path_point,
                             species=conf.species_dir,
                             gene_list=conf.specific_gene,
                             ignore_indels=conf.ignore_indels,
                             ignore_stop_codons=conf.ignore_stop_codons)

        if (conf.inputfasta):

            method = PointFinder.TYPE_BLAST

            blast_run = finder.blast(inputfile=conf.inputfasta,
                                     out_path=conf.outPath_point_blast,
                                     min_cov=0.01,  # Sorts on coverage later
                                     threshold=conf.pf_gene_id,
                                     blast=conf.blast,
                                     cut_off=False)
            results = blast_run.results

        else:
            kma_nano_args = None
            if conf.nanopore:
                kma_nano_args = '-ont -md 5'
            
            method = PointFinder.TYPE_KMA
            
            kma_run = finder.kma(inputfile_1=conf.inputfastq_1,
                                    inputfile_2=conf.inputfastq_2,
                                    out_path=conf.outPath_point_kma,
                                    db_path_kma=conf.db_path_point,
                                    databases=[conf.species_dir],
                                    min_cov=0.01,  # Sorts on coverage later
                                    threshold=conf.pf_gene_id,
                                    kma_path=conf.kma,
                                    sample_name=conf.sample_name,
                                    kma_cge=True,
                                    kma_apm="p",
                                    kma_1t1=True,
                                    kma_add_args=kma_nano_args)

            results = kma_run.results
        
        PointFinderResultHandler.rename_promoter_results(results, method)
        
        if (conf.specific_gene):
            results = PointFinder.discard_from_dict(
                in_dict=results, wanted_list=conf.specific_gene)

        if (method == PointFinder.TYPE_BLAST):
            results_pnt = finder.find_best_seqs(results, conf.pf_gene_cov)
        else:
            results_pnt = results[finder.species]
            if (results_pnt == "No hit found"):
                results_pnt = {}
            else:
                results_pnt["excluded"] = results["excluded"]

        # DEPRECATED
        # mutations in raw reads is only found with write_results.
        # TODO: make a write method that depends on the json output
        finder.write_results(
            out_path=conf.outputPath, result=results, res_type=method,
            unknown_flag=conf.unknown_mut, min_cov=conf.pf_gene_cov,
            perc_iden=conf.pf_gene_id)

        if not conf.unknown_mut:
            results_pnt = PointFinder.discard_unknown_muts(
                results_pnt=results_pnt, phenodb=res_pheno_db, method=method)

        PointFinderResultHandler.standardize_results(std_result,
                                                     results_pnt,
                                                     "PointFinder")

    ##########################################################################
    # Phenotype to genotype
    ##########################################################################

    # Isolate object store results
    isolate = Isolate(name=conf.sample_name, species=conf.species,
                      amr_panel_file=conf.db_panels_file)

    if (conf.acquired or conf.disinf):
        isolate.load_finder_results(std_table=std_result,
                                    phenodb=res_pheno_db,
                                    type="seq_regions")
    if (conf.point):
        isolate.load_finder_results(std_table=std_result,
                                    phenodb=res_pheno_db,
                                    type="seq_variations")

    isolate.calc_res_profile(res_pheno_db)
    ResFinderResultHandler.load_res_profile(std_result, isolate,
                                            conf.amr_abbreviations)

    if (conf.out_json):
        std_result_file = conf.out_json
    else:
        std_result_file = "{}/{}.json".format(
            conf.outputPath, conf.sample_name.replace("_R1", "").split(".")[0])
    with open(std_result_file, 'w') as fh:
        fh.write(std_result.json_dumps())

    # Create and write the downloadable tab file
    pheno_profile_str = isolate.profile_to_str_table(header=True)

    # TODO: REMOVE THE NEED FOR THE PICKLED FILE
    if (conf.pickle):
        isolate_pickle = open("{}/isolate.p".format(conf.outputPath), "wb")
        pickle.dump(isolate, isolate_pickle, protocol=2)

    pheno_table_file = "{}/pheno_table.txt".format(conf.outputPath)
    with open(pheno_table_file, 'w') as fh:
        fh.write(pheno_profile_str)

    if (conf.species is not None):
        # Apply AMR panel
        input_amr_panels = "{}/phenotype_panels.txt".format(conf.db_path_res)
        res_sum_table = ResSumTable(pheno_profile_str)
        res_sum_table.load_amr_panels(input_amr_panels)

        try:
            panel_profile_str = res_sum_table.get_amr_panel_str(
                panel_name_raw=conf.species, header=True)
        # If specified species does not have an associated panel, just ignore it
        # and exit.
        except PanelNameError:
            if conf.species != 'other':
                eprint("Warning: No panel was detected for the species: {}"
                       .format(conf.species))
            sys.exit()

        amr_panel_filename = conf.species.replace(" ", "_")

        panel_tabel_file = ("{tablename}_{species}.txt"
                            .format(tablename=pheno_table_file[:-4],
                                    species=amr_panel_filename))

        with open(panel_tabel_file, "w") as fh:
            fh.write(panel_profile_str)

    return 0


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3


from ..phenotype2genotype.feature import ResGene, ResMutation


class PhenotypeResult(dict):
    def __init__(self, antibiotic, isolate):
        self["type"] = "phenotype"
        self["category"] = "amr"
        self["key"] = antibiotic.name
        self["amr_classes"] = antibiotic.classes
        self["amr_resistance"] = antibiotic.name
        self["amr_resistant"] = False
        self["amr_species_relevant"] = PhenotypeResult.get_amr_relevance(
            antibiotic.name, isolate)
        self["grade"] = 0

    @staticmethod
    def get_amr_relevance(ab, isolate):
        if(isolate.amr_panel is None):
            return True
        elif(ab in isolate.amr_panel):
            return True
        else:
            return False

    def set_resistant(self, res):
        self["amr_resistant"] = res

    def add_feature(self, res_collection, isolate, feature):
        """
            Input:
                res_collection: Result object created by the cgelib package.
                isolate: Isolate object. Must have been loaded with at
                         resistance profile using Isolate.load_finder_results
                         and then Isolate.calc_res_profile.
                feature: Either ResGene or ResMutation object (inherit feature)
        """
        # Get all keys in the result that matches the feature in question.
        # Most of the time this will be a one to one relationship.
        # However if several identical features has been found in a sample,
        # they will all have different keys, but identical ref ids.
        ref_id, type = PhenotypeResult.get_ref_id_and_type(feature, isolate)
        feature_keys = PhenotypeResult.get_keys_matching_ref_id(
            ref_id, res_collection[type])

        # Add keys to phenotype results
        pheno_feat_keys = self.get(type, [])
        for feature_key in feature_keys:
            if feature_key not in pheno_feat_keys:
                pheno_feat_keys.append(feature_key)
        self[type] = pheno_feat_keys

        # Add phenotype keys to feature results
        features = res_collection[type]
        for feat_key in feature_keys:
            feat_result = features[feat_key]
            pheno_keys = feat_result.get("phenotypes", [])
            if self["key"] not in pheno_keys:
                pheno_keys.append(self["key"])
            feat_result["phenotypes"] = pheno_keys

            # Compare the grade of the feature with what is already recorded,
            # keep the greater value.
            # seq_variations has not grade value as it will always be 3
            feat_grade = feat_result.get("grade", 3)
            if feat_grade > self["grade"]:
                self["grade"] = feat_grade


        # Add unique PMIDs to feature results
        if(feature.pmids is not None):
            for pmid in feature.pmids:
                if pmid not in feat_result["pmids"]:
                    feat_result["pmids"].append(pmid)

        # Add unique Notes to feature results
        if(feature.notes is not None):
            if feature.notes not in feat_result["notes"]:
                feat_result["notes"].append(feature.notes)
        db_name = feature.ref_db
        if(type == "seq_regions"):
            db_key = res_collection.get_db_key(db_name)[0]
        elif(type == "seq_variations"):
            db_key = res_collection.get_db_key(db_name)[0]
        self["ref_database"] = [db_key]

    @staticmethod
    def get_ref_id_and_type(feature, isolate):
        """
            Input:
                feature: Either ResGene or ResMutation object (inherit feature)
                isolate: Isolate object. Must have been loaded with at
                         resistance profile using Isolate.load_finder_results
                         and then Isolate.calc_res_profile.
            Output (tuple):
                ref_id: id to identity the feature in relevant database
                type: 'seq_regions' or 'seq_variations'
        """
        type = None
        ref_id = None
        if(isinstance(feature, ResGene)):
            type = "seq_regions"
            ref_id = isolate.resprofile.phenodb.id_to_idwithvar[
                feature.unique_id]
        elif(isinstance(feature, ResMutation)):
            type = "seq_variations"
            if feature.nuc_format == '':
                ref_id = feature.aa_format
            else:
                ref_id = feature.nuc_format
        return (ref_id, type)

    @staticmethod
    def get_keys_matching_ref_id(ref_id, res_collection):
        """
            Input:
                ref_id: Key/ID to search for
                res_collection: Result object created by the cgelib package.
                                The Result object used with this method is
                                found within the main Result object and is
                                accessed either by main_result["seq_regions"]
                                or by main_result["seq_variations"].
            Output (list):
                out_keys: Returns a list of keys where the
                          input ref_id == res_collection[key]["ref_id"].
                          No matches will return an empty list.
        """
        out_keys = []
        for key, results in res_collection.items():
            if(ref_id == results["ref_id"]):
                out_keys.append(key)

        return out_keys

    def get_pmid_key(self, isolate):
        phenodb = isolate.resprofile.phenodb
        isolate_ab = isolate.resprofile.resistance[self["key"]]
        pubmed_ids = isolate_ab.get_pubmed_ids(phenodb)
        return pubmed_ids



#!/usr/bin/env python3


class ResFinderError(Exception):
    """ Root for all module exceptions in the resfinder application.
        Only used to enable "except" of all exceptions from resfinder.
        Never raised.
    """
    pass


class DatabaseError(ResFinderError):
    """ Raise when adding the same key to a dict twice."""
    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(DatabaseError, self).__init__(message, *args)


class DuplicateKeyError(ResFinderError):
    """ Raise when adding the same key to a dict twice."""
    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(DuplicateKeyError, self).__init__(message, *args)


class LockedObjectError(ResFinderError):
    """ Raise when attempting to alter a locked object."""
    def __init__(self, message, *args):
        self.message = message
        # allow users initialize misc. arguments as any other builtin Error
        super(LockedObjectError, self).__init__(message, *args)
